@@ -20,12 +20,17 @@
 package virt_api
 
 import (
+	"context"
 	"crypto/tls"
 	"encoding/json"
 	"fmt"
 	"io/ioutil"
 	"net/http"
+	"os"
+	"os/signal"
 	"sync"
+	"syscall"
+	"time"
 
 	restful "github.com/emicklei/go-restful"
 	"github.com/go-openapi/spec"
@@ -35,8 +40,11 @@ import (
 	"k8s.io/apimachinery/pkg/runtime/schema"
 	"k8s.io/client-go/tools/cache"
 	certificate2 "k8s.io/client-go/util/certificate"
+	"k8s.io/client-go/util/flowcontrol"
 	aggregatorclient "k8s.io/kube-aggregator/pkg/client/clientset_generated/clientset"
 
+	"kubevirt.io/kubevirt/pkg/util/ratelimiter"
+
 	v1 "kubevirt.io/client-go/api/v1"
 	"kubevirt.io/client-go/kubecli"
 	"kubevirt.io/client-go/log"
@@ -55,7 +63,8 @@ import (
 	mutating_webhook "kubevirt.io/kubevirt/pkg/virt-api/webhooks/mutating-webhook"
 	validating_webhook "kubevirt.io/kubevirt/pkg/virt-api/webhooks/validating-webhook"
 	virtconfig "kubevirt.io/kubevirt/pkg/virt-config"
-	"kubevirt.io/kubevirt/pkg/virt-operator/creation/components"
+	"kubevirt.io/kubevirt/pkg/virt-operator/resource/generate/components"
+	virtoperatorutils "kubevirt.io/kubevirt/pkg/virt-operator/util"
 )
 
 const (
@@ -72,6 +81,10 @@ const (
 	defaultTlsKeyFilePath      = "/etc/virt-api/certificates/tls.key"
 	defaultHandlerCertFilePath = "/etc/virt-handler/clientcertificates/tls.crt"
 	defaultHandlerKeyFilePath  = "/etc/virt-handler/clientcertificates/tls.key"
+
+	httpStatusNotFoundMessage     = "Not Found"
+	httpStatusBadRequestMessage   = "Bad Request"
+	httpStatusInternalServerError = "Internal Server Error"
 )
 
 type VirtApi interface {
@@ -93,6 +106,7 @@ type virtAPIApp struct {
 	clusterConfig    *virtconfig.ClusterConfig
 
 	namespace               string
+	host                    string
 	tlsConfig               *tls.Config
 	certificate             *tls.Certificate
 	consoleServerPort       int
@@ -100,15 +114,20 @@ type virtAPIApp struct {
 	handlerTLSConfiguration *tls.Config
 	handlerCertManager      certificate2.Manager
 
-	caConfigMapName     string
-	tlsCertFilePath     string
-	tlsKeyFilePath      string
-	handlerCertFilePath string
-	handlerKeyFilePath  string
-	externallyManaged   bool
+	caConfigMapName              string
+	tlsCertFilePath              string
+	tlsKeyFilePath               string
+	handlerCertFilePath          string
+	handlerKeyFilePath           string
+	externallyManaged            bool
+	reloadableRateLimiter        *ratelimiter.ReloadableRateLimiter
+	reloadableWebhookRateLimiter *ratelimiter.ReloadableRateLimiter
 }
 
-var _ service.Service = &virtAPIApp{}
+var (
+	_                service.Service = &virtAPIApp{}
+	apiHealthVersion                 = new(healthz.KubeApiHealthzVersion)
+)
 
 func NewVirtApi() VirtApi {
 
@@ -120,27 +139,28 @@ func NewVirtApi() VirtApi {
 }
 
 func (app *virtAPIApp) Execute() {
-	virtCli, err := kubecli.GetKubevirtClient()
+	app.reloadableRateLimiter = ratelimiter.NewReloadableRateLimiter(flowcontrol.NewTokenBucketRateLimiter(virtconfig.DefaultVirtAPIQPS, virtconfig.DefaultVirtAPIBurst))
+	app.reloadableWebhookRateLimiter = ratelimiter.NewReloadableRateLimiter(flowcontrol.NewTokenBucketRateLimiter(virtconfig.DefaultVirtWebhookClientQPS, virtconfig.DefaultVirtWebhookClientBurst))
+
+	clientConfig, err := kubecli.GetKubevirtClientConfig()
+	if err != nil {
+		panic(err)
+	}
+	clientConfig.RateLimiter = app.reloadableRateLimiter
+	app.virtCli, err = kubecli.GetKubevirtClientFromRESTConfig(clientConfig)
 	if err != nil {
 		panic(err)
 	}
 
-	authorizor, err := rest.NewAuthorizor()
+	authorizor, err := rest.NewAuthorizor(app.reloadableWebhookRateLimiter)
 	if err != nil {
 		panic(err)
 	}
 
-	config, err := kubecli.GetConfig()
-	if err != nil {
-		panic(err)
-	}
-
-	app.aggregatorClient = aggregatorclient.NewForConfigOrDie(config)
+	app.aggregatorClient = aggregatorclient.NewForConfigOrDie(clientConfig)
 
 	app.authorizor = authorizor
 
-	app.virtCli = virtCli
-
 	app.certsDirectory, err = ioutil.TempDir("", "certsdir")
 	if err != nil {
 		panic(err)
@@ -195,87 +215,148 @@ func (app *virtAPIApp) composeSubresources() {
 			To(subresourceApp.RestartVMRequestHandler).
 			Reads(v1.RestartOptions{}).
 			Param(rest.NamespaceParam(subws)).Param(rest.NameParam(subws)).
-			Operation("restart").
+			Operation(version.Version+"Restart").
 			Doc("Restart a VirtualMachine object.").
 			Returns(http.StatusOK, "OK", "").
-			Returns(http.StatusNotFound, "Not Found", "").
-			Returns(http.StatusBadRequest, "Bad Request", "")
+			Returns(http.StatusNotFound, httpStatusNotFoundMessage, "").
+			Returns(http.StatusBadRequest, httpStatusBadRequestMessage, "")
 		restartRouteBuilder.ParameterNamed("body").Required(false)
 		subws.Route(restartRouteBuilder)
 
 		subws.Route(subws.PUT(rest.ResourcePath(subresourcesvmGVR)+rest.SubResourcePath("migrate")).
 			To(subresourceApp.MigrateVMRequestHandler).
 			Param(rest.NamespaceParam(subws)).Param(rest.NameParam(subws)).
-			Operation("migrate").
+			Operation(version.Version+"Migrate").
 			Doc("Migrate a running VirtualMachine to another node.").
 			Returns(http.StatusOK, "OK", "").
-			Returns(http.StatusNotFound, "Not Found", "").
-			Returns(http.StatusBadRequest, "Bad Request", ""))
+			Returns(http.StatusNotFound, httpStatusNotFoundMessage, "").
+			Returns(http.StatusBadRequest, httpStatusBadRequestMessage, ""))
 
 		subws.Route(subws.PUT(rest.ResourcePath(subresourcesvmGVR)+rest.SubResourcePath("start")).
 			To(subresourceApp.StartVMRequestHandler).
 			Param(rest.NamespaceParam(subws)).Param(rest.NameParam(subws)).
-			Operation("start").
+			Operation(version.Version+"Start").
 			Doc("Start a VirtualMachine object.").
 			Returns(http.StatusOK, "OK", "").
-			Returns(http.StatusNotFound, "Not Found", "").
-			Returns(http.StatusBadRequest, "Bad Request", ""))
+			Returns(http.StatusNotFound, httpStatusNotFoundMessage, "").
+			Returns(http.StatusBadRequest, httpStatusBadRequestMessage, ""))
 
-		subws.Route(subws.PUT(rest.ResourcePath(subresourcesvmGVR)+rest.SubResourcePath("stop")).
+		stopRouteBuilder := subws.PUT(rest.ResourcePath(subresourcesvmGVR)+rest.SubResourcePath("stop")).
 			To(subresourceApp.StopVMRequestHandler).
+			Reads(v1.StopOptions{}).
 			Param(rest.NamespaceParam(subws)).Param(rest.NameParam(subws)).
-			Operation("stop").
+			Operation(version.Version+"Stop").
 			Doc("Stop a VirtualMachine object.").
 			Returns(http.StatusOK, "OK", "").
-			Returns(http.StatusNotFound, "Not Found", "").
-			Returns(http.StatusBadRequest, "Bad Request", ""))
+			Returns(http.StatusNotFound, httpStatusNotFoundMessage, "").
+			Returns(http.StatusBadRequest, httpStatusBadRequestMessage, "")
+		stopRouteBuilder.ParameterNamed("body").Required(false)
+		subws.Route(stopRouteBuilder)
+
+		subws.Route(subws.PUT(rest.ResourcePath(subresourcesvmiGVR)+rest.SubResourcePath("freeze")).
+			To(subresourceApp.FreezeVMIRequestHandler).
+			Param(rest.NamespaceParam(subws)).Param(rest.NameParam(subws)).
+			Operation(version.Version+"Freeze").
+			Doc("Freeze a VirtualMachineInstance object.").
+			Returns(http.StatusOK, "OK", "").
+			Returns(http.StatusInternalServerError, httpStatusInternalServerError, ""))
+
+		subws.Route(subws.PUT(rest.ResourcePath(subresourcesvmiGVR)+rest.SubResourcePath("unfreeze")).
+			To(subresourceApp.UnfreezeVMIRequestHandler).
+			Param(rest.NamespaceParam(subws)).Param(rest.NameParam(subws)).
+			Operation(version.Version+"Unfreeze").
+			Doc("Unfreeze a VirtualMachineInstance object.").
+			Returns(http.StatusOK, "OK", "").
+			Returns(http.StatusInternalServerError, httpStatusInternalServerError, ""))
 
 		subws.Route(subws.PUT(rest.ResourcePath(subresourcesvmiGVR)+rest.SubResourcePath("pause")).
 			To(subresourceApp.PauseVMIRequestHandler).
 			Param(rest.NamespaceParam(subws)).Param(rest.NameParam(subws)).
-			Operation("pause").
+			Operation(version.Version+"Pause").
 			Doc("Pause a VirtualMachineInstance object.").
 			Returns(http.StatusOK, "OK", "").
-			Returns(http.StatusNotFound, "Not Found", "").
-			Returns(http.StatusBadRequest, "Bad Request", ""))
+			Returns(http.StatusNotFound, httpStatusNotFoundMessage, "").
+			Returns(http.StatusBadRequest, httpStatusBadRequestMessage, ""))
 
 		subws.Route(subws.PUT(rest.ResourcePath(subresourcesvmiGVR)+rest.SubResourcePath("unpause")).
 			To(subresourceApp.UnpauseVMIRequestHandler). // handles VMIs as well
 			Param(rest.NamespaceParam(subws)).Param(rest.NameParam(subws)).
-			Operation("unpause").
+			Operation(version.Version+"Unpause").
 			Doc("Unpause a VirtualMachineInstance object.").
 			Returns(http.StatusOK, "OK", "").
-			Returns(http.StatusNotFound, "Not Found", "").
-			Returns(http.StatusBadRequest, "Bad Request", ""))
+			Returns(http.StatusNotFound, httpStatusNotFoundMessage, "").
+			Returns(http.StatusBadRequest, httpStatusBadRequestMessage, ""))
 
 		subws.Route(subws.GET(rest.ResourcePath(subresourcesvmiGVR) + rest.SubResourcePath("console")).
 			To(subresourceApp.ConsoleRequestHandler).
 			Param(rest.NamespaceParam(subws)).Param(rest.NameParam(subws)).
-			Operation("console").
+			Operation(version.Version + "Console").
 			Doc("Open a websocket connection to a serial console on the specified VirtualMachineInstance."))
 
 		subws.Route(subws.GET(rest.ResourcePath(subresourcesvmiGVR) + rest.SubResourcePath("vnc")).
 			To(subresourceApp.VNCRequestHandler).
 			Param(rest.NamespaceParam(subws)).Param(rest.NameParam(subws)).
-			Operation("vnc").
+			Operation(version.Version + "VNC").
 			Doc("Open a websocket connection to connect to VNC on the specified VirtualMachineInstance."))
 
+		subws.Route(subws.GET(rest.ResourcePath(subresourcesvmiGVR) + rest.SubResourcePath("usbredir")).
+			To(subresourceApp.USBRedirRequestHandler).
+			Param(rest.NamespaceParam(subws)).
+			Param(rest.NameParam(subws)).
+			Operation(version.Version + "usbredir").
+			Doc("Open a websocket connection to connect to USB device on the specified VirtualMachineInstance."))
+
+		// VMI endpoint
+		subws.Route(subws.GET(rest.ResourcePath(subresourcesvmiGVR) + rest.SubResourcePath("portforward") + rest.PortPath).
+			To(subresourceApp.PortForwardRequestHandler(subresourceApp.FetchVirtualMachineInstance)).
+			Param(rest.NamespaceParam(subws)).Param(rest.NameParam(subws)).
+			Param(rest.PortForwardPortParameter(subws)).
+			Operation(version.Version + "vmi-PortForward").
+			Doc("Open a websocket connection forwarding traffic to the specified VirtualMachineInstance and port."))
+		subws.Route(subws.GET(rest.ResourcePath(subresourcesvmiGVR) + rest.SubResourcePath("portforward") + rest.PortPath + rest.ProtocolPath).
+			To(subresourceApp.PortForwardRequestHandler(subresourceApp.FetchVirtualMachineInstance)).
+			Param(rest.NamespaceParam(subws)).Param(rest.NameParam(subws)).
+			Param(rest.PortForwardPortParameter(subws)).
+			Param(rest.PortForwardProtocolParameter(subws)).
+			Operation(version.Version + "vmi-PortForwardWithProtocol").
+			Doc("Open a websocket connection forwarding traffic of the specified protocol (either tcp or udp) to the specified VirtualMachineInstance and port."))
+
+		// VM endpoint
+		subws.Route(subws.GET(rest.ResourcePath(subresourcesvmGVR) + rest.SubResourcePath("portforward") + rest.PortPath).
+			To(subresourceApp.PortForwardRequestHandler(subresourceApp.FetchVirtualMachineInstanceForVM)).
+			Param(rest.NamespaceParam(subws)).Param(rest.NameParam(subws)).
+			Param(rest.PortForwardPortParameter(subws)).
+			Operation(version.Version + "vm-PortForward").
+			Doc("Open a websocket connection forwarding traffic to the running VMI for the specified VirtualMachine and port."))
+		subws.Route(subws.GET(rest.ResourcePath(subresourcesvmGVR) + rest.SubResourcePath("portforward") + rest.PortPath + rest.ProtocolPath).
+			To(subresourceApp.PortForwardRequestHandler(subresourceApp.FetchVirtualMachineInstanceForVM)).
+			Param(rest.NamespaceParam(subws)).Param(rest.NameParam(subws)).
+			Param(rest.PortForwardPortParameter(subws)).
+			Param(rest.PortForwardProtocolParameter(subws)).
+			Operation(version.Version + "vm-PortForwardWithProtocol").
+			Doc("Open a websocket connection forwarding traffic of the specified protocol (either tcp or udp) to the specified VirtualMachine and port."))
+
 		// An empty handler function would respond with HTTP OK by default
 		subws.Route(subws.GET(rest.ResourcePath(subresourcesvmiGVR) + rest.SubResourcePath("test")).
 			To(func(request *restful.Request, response *restful.Response) {}).
 			Param(rest.NamespaceParam(subws)).Param(rest.NameParam(subws)).
-			Operation("test").
+			Operation(version.Version + "Test").
 			Doc("Test endpoint verifying apiserver connectivity."))
 
 		subws.Route(subws.GET(rest.SubResourcePath("version")).Produces(restful.MIME_JSON).
 			To(func(request *restful.Request, response *restful.Response) {
 				response.WriteAsJson(virtversion.Get())
-			}).Operation("version"))
+			}).Operation(version.Version + "Version"))
+		subws.Route(subws.GET(rest.SubResourcePath("guestfs")).Produces(restful.MIME_JSON).
+			To(app.GetGsInfo()).
+			Operation(version.Version+"Guestfs").
+			Returns(http.StatusOK, "OK", "").
+			Returns(http.StatusBadRequest, httpStatusBadRequestMessage, ""))
 		subws.Route(subws.GET(rest.SubResourcePath("healthz")).
-			To(healthz.KubeConnectionHealthzFuncFactory(app.clusterConfig)).
+			To(healthz.KubeConnectionHealthzFuncFactory(app.clusterConfig, apiHealthVersion)).
 			Consumes(restful.MIME_JSON).
 			Produces(restful.MIME_JSON).
-			Operation("checkHealth").
+			Operation(version.Version+"CheckHealth").
 			Doc("Health endpoint").
 			Returns(http.StatusOK, "OK", "").
 			Returns(http.StatusInternalServerError, "Unhealthy", ""))
@@ -284,26 +365,16 @@ func (app *virtAPIApp) composeSubresources() {
 			Param(rest.NamespaceParam(subws)).Param(rest.NameParam(subws)).
 			Consumes(restful.MIME_JSON).
 			Produces(restful.MIME_JSON).
-			Operation("guestosinfo").
+			Operation(version.Version+"Guestosinfo").
 			Doc("Get guest agent os information").
 			Writes(v1.VirtualMachineInstanceGuestAgentInfo{}).
 			Returns(http.StatusOK, "OK", v1.VirtualMachineInstanceGuestAgentInfo{}))
 
-		subws.Route(subws.PUT(rest.ResourcePath(subresourcesvmGVR)+rest.SubResourcePath("rename")).
-			To(subresourceApp.RenameVMRequestHandler).
-			Param(rest.NamespaceParam(subws)).Param(rest.NameParam(subws)).
-			Operation("rename").
-			Doc("Rename a stopped VirtualMachine object.").
-			Returns(http.StatusOK, "OK", "").
-			Returns(http.StatusAccepted, "Accepted", "").
-			Returns(http.StatusNotFound, "Not Found", "").
-			Returns(http.StatusBadRequest, "Bad Request", ""))
-
 		subws.Route(subws.GET(rest.ResourcePath(subresourcesvmiGVR)+rest.SubResourcePath("userlist")).
 			To(subresourceApp.UserList).
 			Consumes(restful.MIME_JSON).
 			Produces(restful.MIME_JSON).
-			Operation("userlist").
+			Operation(version.Version+"Userlist").
 			Doc("Get list of active users via guest agent").
 			Writes(v1.VirtualMachineInstanceGuestOSUserList{}).
 			Returns(http.StatusOK, "OK", v1.VirtualMachineInstanceGuestOSUserList{}))
@@ -312,42 +383,46 @@ func (app *virtAPIApp) composeSubresources() {
 			To(subresourceApp.FilesystemList).
 			Consumes(restful.MIME_JSON).
 			Produces(restful.MIME_JSON).
-			Operation("filesystemlist").
+			Operation(version.Version+"Filesystemlist").
 			Doc("Get list of active filesystems on guest machine via guest agent").
 			Writes(v1.VirtualMachineInstanceFileSystemList{}).
 			Returns(http.StatusOK, "OK", v1.VirtualMachineInstanceFileSystemList{}))
 
 		subws.Route(subws.PUT(rest.ResourcePath(subresourcesvmiGVR)+rest.SubResourcePath("addvolume")).
 			To(subresourceApp.VMIAddVolumeRequestHandler).
+			Reads(v1.AddVolumeOptions{}).
 			Param(rest.NamespaceParam(subws)).Param(rest.NameParam(subws)).
-			Operation("vmi-addvolume").
+			Operation(version.Version+"vmi-addvolume").
 			Doc("Add a volume and disk to a running Virtual Machine Instance").
 			Returns(http.StatusOK, "OK", "").
-			Returns(http.StatusBadRequest, "Bad Request", ""))
+			Returns(http.StatusBadRequest, httpStatusBadRequestMessage, ""))
 
 		subws.Route(subws.PUT(rest.ResourcePath(subresourcesvmiGVR)+rest.SubResourcePath("removevolume")).
 			To(subresourceApp.VMIRemoveVolumeRequestHandler).
+			Reads(v1.RemoveVolumeOptions{}).
 			Param(rest.NamespaceParam(subws)).Param(rest.NameParam(subws)).
-			Operation("vmi-removevolume").
+			Operation(version.Version+"vmi-removevolume").
 			Doc("Removes a volume and disk from a running Virtual Machine Instance").
 			Returns(http.StatusOK, "OK", "").
-			Returns(http.StatusBadRequest, "Bad Request", ""))
+			Returns(http.StatusBadRequest, httpStatusBadRequestMessage, ""))
 
 		subws.Route(subws.PUT(rest.ResourcePath(subresourcesvmGVR)+rest.SubResourcePath("addvolume")).
 			To(subresourceApp.VMAddVolumeRequestHandler).
+			Reads(v1.AddVolumeOptions{}).
 			Param(rest.NamespaceParam(subws)).Param(rest.NameParam(subws)).
-			Operation("vm-addvolume").
+			Operation(version.Version+"vm-addvolume").
 			Doc("Add a volume and disk to a running Virtual Machine.").
 			Returns(http.StatusOK, "OK", "").
-			Returns(http.StatusBadRequest, "Bad Request", ""))
+			Returns(http.StatusBadRequest, httpStatusBadRequestMessage, ""))
 
 		subws.Route(subws.PUT(rest.ResourcePath(subresourcesvmGVR)+rest.SubResourcePath("removevolume")).
 			To(subresourceApp.VMRemoveVolumeRequestHandler).
+			Reads(v1.RemoveVolumeOptions{}).
 			Param(rest.NamespaceParam(subws)).Param(rest.NameParam(subws)).
-			Operation("vm-removevolume").
+			Operation(version.Version+"vm-removevolume").
 			Doc("Removes a volume and disk from a running Virtual Machine.").
 			Returns(http.StatusOK, "OK", "").
-			Returns(http.StatusBadRequest, "Bad Request", ""))
+			Returns(http.StatusBadRequest, httpStatusBadRequestMessage, ""))
 
 		// Return empty api resource list.
 		// K8s expects to be able to retrieve a resource list for each aggregated
@@ -371,6 +446,10 @@ func (app *virtAPIApp) composeSubresources() {
 						Name:       "virtualmachineinstances/console",
 						Namespaced: true,
 					},
+					{
+						Name:       "virtualmachineinstances/portforward",
+						Namespaced: true,
+					},
 					{
 						Name:       "virtualmachineinstances/pause",
 						Namespaced: true,
@@ -379,6 +458,14 @@ func (app *virtAPIApp) composeSubresources() {
 						Name:       "virtualmachineinstances/unpause",
 						Namespaced: true,
 					},
+					{
+						Name:       "virtualmachineinstances/freeze",
+						Namespaced: true,
+					},
+					{
+						Name:       "virtualmachineinstances/unfreeze",
+						Namespaced: true,
+					},
 					{
 						Name:       "virtualmachines/start",
 						Namespaced: true,
@@ -399,10 +486,6 @@ func (app *virtAPIApp) composeSubresources() {
 						Name:       "virtualmachineinstances/guestosinfo",
 						Namespaced: true,
 					},
-					{
-						Name:       "virtualmachines/rename",
-						Namespaced: true,
-					},
 					{
 						Name:       "virtualmachineinstances/userlist",
 						Namespaced: true,
@@ -423,10 +506,10 @@ func (app *virtAPIApp) composeSubresources() {
 
 				response.WriteAsJson(list)
 			}).
-			Operation("getAPISubResources").
+			Operation(version.Version+"getAPISubResources").
 			Doc("Get a KubeVirt API resources").
 			Returns(http.StatusOK, "OK", metav1.APIResourceList{}).
-			Returns(http.StatusNotFound, "Not Found", ""))
+			Returns(http.StatusNotFound, httpStatusNotFoundMessage, ""))
 
 		restful.Add(subws)
 
@@ -453,8 +536,8 @@ func (app *virtAPIApp) composeSubresources() {
 		Operation("getRootPaths").
 		Doc("Get KubeVirt API root paths").
 		Returns(http.StatusOK, "OK", metav1.RootPaths{}).
-		Returns(http.StatusNotFound, "Not Found", ""))
-	ws.Route(ws.GET("/healthz").To(healthz.KubeConnectionHealthzFuncFactory(app.clusterConfig)).Doc("Health endpoint"))
+		Returns(http.StatusNotFound, httpStatusNotFoundMessage, ""))
+	ws.Route(ws.GET("/healthz").To(healthz.KubeConnectionHealthzFuncFactory(app.clusterConfig, apiHealthVersion)).Doc("Health endpoint"))
 
 	for _, version := range v1.SubresourceGroupVersions {
 		// K8s needs the ability to query info about a specific API group
@@ -463,10 +546,10 @@ func (app *virtAPIApp) composeSubresources() {
 			To(func(request *restful.Request, response *restful.Response) {
 				response.WriteAsJson(subresourceAPIGroup())
 			}).
-			Operation("getSubAPIGroup").
+			Operation(version.Version+"GetSubAPIGroup").
 			Doc("Get a KubeVirt API Group").
 			Returns(http.StatusOK, "OK", metav1.APIGroup{}).
-			Returns(http.StatusNotFound, "Not Found", ""))
+			Returns(http.StatusNotFound, httpStatusNotFoundMessage, ""))
 	}
 
 	// K8s needs the ability to query the list of API groups this endpoint supports
@@ -481,7 +564,7 @@ func (app *virtAPIApp) composeSubresources() {
 		Operation("getAPIGroupList").
 		Doc("Get a KubeVirt API GroupList").
 		Returns(http.StatusOK, "OK", metav1.APIGroupList{}).
-		Returns(http.StatusNotFound, "Not Found", ""))
+		Returns(http.StatusNotFound, httpStatusNotFoundMessage, ""))
 
 	once := sync.Once{}
 	var openapispec *spec.Swagger
@@ -548,7 +631,7 @@ func deserializeStrings(in string) ([]string, error) {
 }
 
 func (app *virtAPIApp) readRequestHeader() error {
-	authConfigMap, err := app.virtCli.CoreV1().ConfigMaps(metav1.NamespaceSystem).Get(util.ExtensionAPIServerAuthenticationConfigMap, metav1.GetOptions{})
+	authConfigMap, err := app.virtCli.CoreV1().ConfigMaps(metav1.NamespaceSystem).Get(context.Background(), util.ExtensionAPIServerAuthenticationConfigMap, metav1.GetOptions{})
 	if err != nil {
 		return err
 	}
@@ -667,9 +750,17 @@ func (app *virtAPIApp) setupTLS(k8sCAManager webhooksutils.ClientCAManager, kube
 	app.handlerTLSConfiguration = webhooksutils.SetupTLSForVirtHandlerClients(kubevirtCAManager, app.handlerCertManager, app.externallyManaged)
 }
 
-func (app *virtAPIApp) startTLS(informerFactory controller.KubeInformerFactory, stopCh <-chan struct{}) error {
+func (app *virtAPIApp) startTLS(informerFactory controller.KubeInformerFactory) error {
 
 	errors := make(chan error)
+	c := make(chan os.Signal, 1)
+
+	signal.Notify(c, os.Interrupt,
+		syscall.SIGHUP,
+		syscall.SIGINT,
+		syscall.SIGTERM,
+		syscall.SIGQUIT,
+	)
 
 	authConfigMapInformer := informerFactory.ApiAuthConfigMap()
 	kubevirtCAConfigInformer := informerFactory.KubeVirtCAConfigMap()
@@ -681,25 +772,64 @@ func (app *virtAPIApp) startTLS(informerFactory controller.KubeInformerFactory,
 
 	app.Compose()
 
+	http.Handle("/metrics", promhttp.Handler())
+	server := &http.Server{
+		Addr:      fmt.Sprintf("%s:%d", app.BindAddress, app.Port),
+		TLSConfig: app.tlsConfig,
+	}
+
 	// start TLS server
 	go func() {
-		http.Handle("/metrics", promhttp.Handler())
-
-		server := &http.Server{
-			Addr:      fmt.Sprintf("%s:%d", app.BindAddress, app.Port),
-			TLSConfig: app.tlsConfig,
-		}
-
 		errors <- server.ListenAndServeTLS("", "")
 	}()
 
+	// start graceful shutdown handler
+	go func() {
+		s := <-c
+		log.Log.Infof("Received signal %s, initiating graceful shutdown", s.String())
+
+		// pause briefly to ensure the load balancer has had a chance to
+		// remove this endpoint from rotation due to pod.DeletionTimestamp != nil
+		// By pausing, we reduce the chance that the load balancer will attempt to
+		// route new requests to the service after we've started the shutdown
+		// procedure
+		time.Sleep(5 * time.Second)
+
+		// by default, server.Shutdown() waits indefinitely for all existing
+		// connections to close. We need to give this a timeout to ensure the
+		// shutdown will eventually complete.
+		ctx, cancel := context.WithTimeout(context.Background(), 5*time.Second)
+		defer func() {
+			cancel()
+		}()
+
+		// Shutdown means new connections are not permitted and it waits for existing
+		// connections to terminate (up to the context timeout)
+		server.Shutdown(ctx)
+		// Shutdown forces any existing connections that persist after the shutdown
+		// times out to be forced closed.
+		server.Close()
+	}()
+
 	// wait for server to exit
-	return <-errors
+	err := <-errors
+
+	if err != nil && err != http.ErrServerClosed {
+		// ErrServerClosed is an expected error during normal shutdown
+		return err
+	}
+	return nil
 }
 
 func (app *virtAPIApp) Run() {
+	host, err := os.Hostname()
+	if err != nil {
+		panic(fmt.Errorf("unable to get hostname: %v", err))
+	}
+	app.host = host
+
 	// get client Cert
-	err := app.readRequestHeader()
+	err = app.readRequestHeader()
 	if err != nil {
 		panic(err)
 	}
@@ -720,6 +850,12 @@ func (app *virtAPIApp) Run() {
 	kubevirtCAConfigInformer := kubeInformerFactory.KubeVirtCAConfigMap()
 	kubeVirtInformer := kubeInformerFactory.KubeVirt()
 
+	// Wire up health check trigger
+	configMapInformer.SetWatchErrorHandler(func(r *cache.Reflector, err error) {
+		apiHealthVersion.Clear()
+		cache.DefaultWatchErrorHandler(r, err)
+	})
+
 	stopChan := make(chan struct{}, 1)
 	defer close(stopChan)
 	go webhookInformers.VMIInformer.Run(stopChan)
@@ -742,18 +878,40 @@ func (app *virtAPIApp) Run() {
 		configMapInformer.HasSynced)
 
 	app.clusterConfig = virtconfig.NewClusterConfig(configMapInformer, crdInformer, kubeVirtInformer, app.namespace)
+	app.clusterConfig.SetConfigModifiedCallback(app.shouldChangeLogVerbosity)
+	app.clusterConfig.SetConfigModifiedCallback(app.shouldChangeRateLimiter)
 
 	go app.certmanager.Start()
 	go app.handlerCertManager.Start()
 
 	// start TLS server
 	// tls server will only accept connections when fetching a certificate and internal configuration passed once
-	err = app.startTLS(kubeInformerFactory, stopChan)
+	err = app.startTLS(kubeInformerFactory)
 	if err != nil {
 		panic(err)
 	}
 }
 
+// Update virt-api log verbosity on relevant config changes
+func (app *virtAPIApp) shouldChangeLogVerbosity() {
+	verbosity := app.clusterConfig.GetVirtAPIVerbosity(app.host)
+	log.Log.SetVerbosityLevel(int(verbosity))
+	log.Log.V(2).Infof("set log verbosity to %d", verbosity)
+}
+
+// Update virt-handler rate limiter
+func (app *virtAPIApp) shouldChangeRateLimiter() {
+	config := app.clusterConfig.GetConfig()
+	qps := config.APIConfiguration.RestClient.RateLimiter.TokenBucketRateLimiter.QPS
+	burst := config.APIConfiguration.RestClient.RateLimiter.TokenBucketRateLimiter.Burst
+	app.reloadableRateLimiter.Set(flowcontrol.NewTokenBucketRateLimiter(qps, burst))
+	log.Log.V(2).Infof("setting rate limiter for the API to %v QPS and %v Burst", qps, burst)
+	qps = config.WebhookConfiguration.RestClient.RateLimiter.TokenBucketRateLimiter.QPS
+	burst = config.WebhookConfiguration.RestClient.RateLimiter.TokenBucketRateLimiter.Burst
+	app.reloadableWebhookRateLimiter.Set(flowcontrol.NewTokenBucketRateLimiter(qps, burst))
+	log.Log.V(2).Infof("setting rate limiter for webhooks to %v QPS and %v Burst", qps, burst)
+}
+
 func (app *virtAPIApp) AddFlags() {
 	app.InitFlags()
 
@@ -778,3 +936,31 @@ func (app *virtAPIApp) AddFlags() {
 	flag.BoolVar(&app.externallyManaged, "externally-managed", false,
 		"Allow intermediate certificates to be used in building up the chain of trust when certificates are externally managed")
 }
+
+// GetGsInfo returns the libguestfs-tools image information based on the KubeVirt installation in the namespace.
+func (app *virtAPIApp) GetGsInfo() func(_ *restful.Request, response *restful.Response) {
+	return func(_ *restful.Request, response *restful.Response) {
+		var kvConfig virtoperatorutils.KubeVirtDeploymentConfig
+		kv := app.clusterConfig.GetConfigFromKubeVirtCR()
+		if kv == nil {
+			error_guestfs(fmt.Errorf("Failed getting KubeVirt config"), response)
+		}
+		err := json.Unmarshal([]byte(kv.Status.ObservedDeploymentConfig), &kvConfig)
+		if err != nil {
+			error_guestfs(err, response)
+			return
+		}
+		response.WriteAsJson(kubecli.GuestfsInfo{
+			Registry: kv.Status.ObservedKubeVirtRegistry,
+			Tag:      kv.Status.ObservedKubeVirtVersion,
+			Digest:   kvConfig.GsSha,
+		})
+		return
+	}
+}
+
+func error_guestfs(err error, response *restful.Response) {
+	res := map[string]interface{}{}
+	res["guestfs"] = map[string]interface{}{"status": "failed", "error": fmt.Sprintf("%v", err)}
+	response.WriteHeaderAndJson(http.StatusInternalServerError, res, restful.MIME_JSON)
+}
