@@ -23,7 +23,7 @@ import (
 	"encoding/base64"
 	"encoding/json"
 	"fmt"
-	"io/ioutil"
+	"os"
 	"path/filepath"
 	"strings"
 	"sync"
@@ -34,6 +34,7 @@ import (
 	v1 "kubevirt.io/client-go/api/v1"
 	"kubevirt.io/client-go/log"
 	"kubevirt.io/kubevirt/pkg/config"
+	"kubevirt.io/kubevirt/pkg/virt-launcher/virtwrap/agent"
 	"kubevirt.io/kubevirt/pkg/virt-launcher/virtwrap/api"
 	"kubevirt.io/kubevirt/pkg/virt-launcher/virtwrap/cli"
 	domainerrors "kubevirt.io/kubevirt/pkg/virt-launcher/virtwrap/errors"
@@ -161,13 +162,13 @@ func (l *AccessCredentialManager) writeGuestFile(contents string, domName string
 	}
 
 	cmdWriteFile := fmt.Sprintf(`{"execute": "guest-file-write", "arguments": { "handle": %d, "buf-b64": "%s" } }`, openRes.Return, base64Str)
-	output, err = l.virConn.QemuAgentCommand(cmdWriteFile, domName)
+	_, err = l.virConn.QemuAgentCommand(cmdWriteFile, domName)
 	if err != nil {
 		return err
 	}
 
 	cmdCloseFile := fmt.Sprintf(`{"execute": "guest-file-close", "arguments": { "handle": %d } }`, openRes.Return)
-	output, err = l.virConn.QemuAgentCommand(cmdCloseFile, domName)
+	_, err = l.virConn.QemuAgentCommand(cmdCloseFile, domName)
 	if err != nil {
 		return err
 	}
@@ -217,7 +218,7 @@ func (l *AccessCredentialManager) readGuestFile(domName string, filePath string)
 	}
 
 	cmdCloseFile := fmt.Sprintf(`{"execute": "guest-file-close", "arguments": { "handle": %d } }`, openRes.Return)
-	output, err = l.virConn.QemuAgentCommand(cmdCloseFile, domName)
+	_, err = l.virConn.QemuAgentCommand(cmdCloseFile, domName)
 	if err != nil {
 		return contents, err
 	}
@@ -226,59 +227,7 @@ func (l *AccessCredentialManager) readGuestFile(domName string, filePath string)
 }
 
 func (l *AccessCredentialManager) agentGuestExec(domName string, command string, args []string) (string, error) {
-	stdOut := ""
-	argsStr := ""
-	for _, arg := range args {
-		if argsStr == "" {
-			argsStr = fmt.Sprintf("\"%s\"", arg)
-		} else {
-			argsStr = argsStr + fmt.Sprintf(", \"%s\"", arg)
-		}
-	}
-
-	cmdExec := fmt.Sprintf(`{"execute": "guest-exec", "arguments": { "path": "%s", "arg": [ %s ], "capture-output":true } }`, command, argsStr)
-	output, err := l.virConn.QemuAgentCommand(cmdExec, domName)
-	execRes := &execReturn{}
-	err = json.Unmarshal([]byte(output), execRes)
-	if err != nil {
-		return "", err
-	}
-
-	if execRes.Return.Pid <= 0 {
-		return "", fmt.Errorf("Invalid pid [%d] returned from qemu agent during access credential injection: %s", execRes.Return.Pid, output)
-	}
-
-	exited := false
-	exitCode := 0
-	for i := 10; i > 0; i-- {
-		cmdExecStatus := fmt.Sprintf(`{"execute": "guest-exec-status", "arguments": { "pid": %d } }`, execRes.Return.Pid)
-		output, err := l.virConn.QemuAgentCommand(cmdExecStatus, domName)
-		execStatusRes := &execStatusReturn{}
-		err = json.Unmarshal([]byte(output), execStatusRes)
-		if err != nil {
-			return "", err
-		}
-
-		if execStatusRes.Return.Exited {
-			stdOutBytes, err := base64.StdEncoding.DecodeString(execStatusRes.Return.OutData)
-			if err != nil {
-				return "", err
-			}
-			stdOut = string(stdOutBytes)
-			exitCode = execStatusRes.Return.ExitCode
-			exited = true
-			break
-		}
-		time.Sleep(1 * time.Second)
-	}
-
-	if !exited {
-		return "", fmt.Errorf("Timed out waiting for guest pid [%d] for command [%s] to exit", execRes.Return.Pid, command)
-	} else if exitCode != 0 {
-		return stdOut, fmt.Errorf("Non-zero exit code [%d] for guest command [%s] with args [%v]: %s", exitCode, command, args, stdOut)
-	}
-
-	return stdOut, nil
+	return agent.GuestExec(l.virConn, domName, command, args, 10)
 }
 
 // Requires usage of mkdir, chown, chmod
@@ -476,7 +425,11 @@ func (l *AccessCredentialManager) reportAccessCredentialResult(vmi *v1.VirtualMa
 		return nil
 	}
 
-	_, err = util.SetDomainSpecStrWithHooks(l.virConn, vmi, domainSpec)
+	d, err := util.SetDomainSpecStrWithHooks(l.virConn, vmi, domainSpec)
+	if err != nil {
+		return err
+	}
+	defer d.Free()
 	return err
 }
 
@@ -543,13 +496,14 @@ func (l *AccessCredentialManager) watchSecrets(vmi *v1.VirtualMachineInstance) {
 
 		// Step 1. Populate Secrets and filepath Map
 		for _, accessCred := range vmi.Spec.AccessCredentials {
+			accessCred := accessCred
 			secretName := getSecret(&accessCred)
 			if secretName == "" {
 				continue
 			}
 
 			secretDir := getSecretDir(secretName)
-			files, err := ioutil.ReadDir(secretDir)
+			files, err := os.ReadDir(secretDir)
 			if err != nil {
 				// if reading failed, reset reload to true so this change will be retried again
 				reload = true
@@ -575,7 +529,7 @@ func (l *AccessCredentialManager) watchSecrets(vmi *v1.VirtualMachineInstance) {
 						continue
 					}
 
-					pubKeyBytes, err := ioutil.ReadFile(filepath.Join(secretDir, file.Name()))
+					pubKeyBytes, err := os.ReadFile(filepath.Join(secretDir, file.Name()))
 					if err != nil {
 						// if reading failed, reset reload to true so this change will be retried again
 						reload = true
@@ -598,7 +552,7 @@ func (l *AccessCredentialManager) watchSecrets(vmi *v1.VirtualMachineInstance) {
 						continue
 					}
 
-					passwordBytes, err := ioutil.ReadFile(filepath.Join(secretDir, file.Name()))
+					passwordBytes, err := os.ReadFile(filepath.Join(secretDir, file.Name()))
 					if err != nil {
 						// if reading failed, reset reload to true so this change will be retried again
 						reload = true
